import sympy
import itertools
from sympy import Symbol, Piecewise, Number, postorder_traversal, numbered_symbols
from pystencils.simp.liveness_opts import *

atom_symbol_iter = numbered_symbols("atom_")


def three_operand_form(assignments):
    """Transforms list of assignments in three operand form"""

    def atomize(expr, atoms):
        if len(expr.args) == 0:
            return expr

        atom = next(atom_symbol_iter)
        if len(expr.args) == 1:
            atoms.append(Assignment(atom, expr.func(atomize(expr.args[0], atoms))))
            return atom

        if isinstance(expr, Piecewise):
            atoms.append(
                Assignment(
                    atom,
                    Piecewise(*[(atomize(expr.expr, atoms), expr.cond) for expr in expr.args])))
            return atom

        atoms.append(Assignment(atom, expr.func(atomize(expr.args[0], atoms), atomize(expr.args[1], atoms))))

        current_atom = atom
        for i in range(2, len(expr.args)):
            atom = next(atom_symbol_iter)
            atoms.append(Assignment(atom, expr.func(atomize(expr.args[i], atoms), current_atom)))
            current_atom = atom

        return current_atom

    atoms = []
    for eq in assignments:
        new_atoms = []
        atomize(eq.rhs, new_atoms)
        if len(new_atoms) > 0:
            new_atoms[-1] = Assignment(eq.lhs, new_atoms[-1].rhs)
        else:
            new_atoms.append(eq)
        atoms.extend(new_atoms)

    return atoms


def var_to_shmem(eqs, var_count=8):
    if var_count > 8:
        return eqs
    if var_count == 0:
        return copy.copy(eqs)
    for eq in eqs:
        if eq.lhs.name.startswith("shmemslot"):
            return eqs

    usage = get_usage(eqs)
    usage_list = [(s, usage[s]) for s in usage]

    usage_list.sort(key=lambda s: -s[1])

    vars = [Symbol("shmemslot" + str(i)) for i in range(0, var_count)]
    shmem_eqs = []
    for idx, eq in enumerate(eqs):
        shmem_eqs.append(eq.subs([(usage_list[i][0], vars[i]) for i in range(0, var_count)]))
    return shmem_eqs


def shift_fa_eqs(eqs, direction=1):
    def shift_fa(expr, direction):
        if isinstance(expr, Field.Access):
            return expr.neighbor(0, direction)
        if len(expr.args) == 0:
            return expr
        else:
            return expr.func(*[shift_fa(arg, direction) for arg in expr.args])

    new_eqs = []
    for eq in eqs:
        new_eqs.append(shift_fa(eq, direction))
    return new_eqs


def get_steal_list(eqs, shifted_eqs):
    def is_equal_arg(left_arg, right_arg, steal_list, left_def, right_def, verbose=False):

        if verbose: print("is_equal_arg: IN left_arg " + str(left_arg))
        if verbose: print("is_equal_arg: IN right_arg " + str(right_arg))

        if verbose: print("is_equal_arg: SUB left_arg " + str(left_arg))
        if verbose: print("is_equal_arg: SUB left_arg " + str(right_arg))

        if isinstance(left_arg, Number): return left_arg == right_arg
        if isinstance(left_arg, Field.Access): return left_arg == right_arg

        if left_arg not in steal_list: return False

        if verbose: print("is_equal_arg: stolen" + str(steal_list[left_arg]))

        return steal_list[left_arg] == right_arg

    def is_equal_expr(left_expr, right_expr, steal_list, left_def, right_def, verbose=False):

        # print(str(left_expr) + " =?= " + str(right_expr))

        if type(left_expr) != type(right_expr): return False

        if left_expr.func != right_expr.func or len(left_expr.args) != len(right_expr.args):
            return False

        if len(left_expr.args) == 0:
            return is_equal_arg(left_expr, right_expr, steal_list, left_def, right_def, verbose)

        for left_arg_perm in itertools.permutations(left_expr.args):
            equal_args = True
            for idx, left_arg in enumerate(left_arg_perm):
                if not is_equal_arg(left_arg, right_expr.args[idx], steal_list, left_def, right_def,
                                    verbose):
                    equal_args = False
                    break
            if equal_args: return True

        return False

    steal_from_e = {}
    left_def = get_definitions(eqs)
    right_def = get_definitions(shifted_eqs)
    for lidx, asgn_left in enumerate(eqs):
        verbose = False

        for left_subexpr in sympy.postorder_traversal(asgn_left.rhs):
            if isinstance(left_subexpr, sympy.Number) or isinstance(
                    left_subexpr, Field.Access) or isinstance(left_subexpr, Assignment):
                continue
            for ridx, asgn_right in enumerate(shifted_eqs):

                for right_subexpr in sympy.postorder_traversal(asgn_right.rhs):
                    left_arg = left_subexpr
                    right_arg = right_subexpr
                    if isinstance(left_subexpr,
                                  Symbol) and not isinstance(left_subexpr, Field.Access):
                        left_arg = left_def[left_subexpr].rhs
                    if isinstance(right_subexpr,
                                  Symbol) and not isinstance(right_subexpr, Field.Access):
                        right_arg = right_def[right_subexpr].rhs

                    if is_equal_expr(left_arg, right_arg, steal_from_e, left_def, right_def,
                                     verbose):
                        steal_from_e[left_subexpr] = right_subexpr
                        # if verbose:
                        print(str(left_subexpr) + " ==  " + str(right_subexpr))

    return steal_from_e


def find_symbol(eqs, name):
    for eq in eqs:
        if eq.lhs.name == name: return eq.lhs


def find_expr(eqs, expr):
    for idx, eq in enumerate(eqs):
        for sub_expr in postorder_traversal(eq):
            if sub_expr == expr:
                return (idx, sub_expr, eq)


def left_steal(eqs, steal_count=2):
    shifted_eqs = shift_fa_eqs(eqs)
    steal_from_e = get_steal_list(eqs, shifted_eqs)

    usage = get_usage(eqs)
    definitions = get_definitions(eqs)

    def count_nodes_up(node):
        if isinstance(node, Field.Access):
            return 1

        if node in definitions:
            node = definitions[node].rhs

        node_count = 0
        for arg in node.args:
            if not (arg in usage and usage[arg] > 1):
                node_count += count_nodes_up(arg)
        return node_count + 1

    new_eqs = copy.copy(eqs)
    for i in range(0, steal_count):

        scores = [(s, count_nodes_up(s)) for s in steal_from_e if isinstance(s, Symbol)]
        scores.sort(key=lambda s: s[1], reverse=True)

        print(scores[0:10])

        sym_xi = scores[0][0]
        print(sym_xi)
        print(steal_from_e[sym_xi])
        steal_src = find_expr(new_eqs, shift_fa_eqs([steal_from_e[sym_xi]], -1)[0])
        shmem_var = Symbol("shmemslot" + str(i))

        new_eqs.insert(steal_src[0] + 1, Assignment(shmem_var, steal_src[1]))

        steal_dst = find_expr(new_eqs, sym_xi)

        print(steal_dst)
        print(steal_src)
        print()

        for idx, eq in enumerate(new_eqs):
            if steal_dst[1] in eq.atoms():
                new_eqs[idx] = Assignment(new_eqs[idx].lhs, new_eqs[idx].rhs.subs(
                    steal_dst[1], shmem_var))

        new_eqs.pop(steal_dst[0])

        # Ancestors of donated value cannot be stolen, therefore remove from steal list
        def get_ancestor_nodes(node, definitions):
            ancestors = [node]
            if isinstance(node, sympy.Number):
                return []
            if node in definitions:
                ancestors.extend(get_ancestor_nodes(definitions[node].rhs, definitions))
            for arg in node.args:
                ancestors.extend(get_ancestor_nodes(arg, definitions))
            return ancestors

        for a in get_ancestor_nodes(steal_from_e[sym_xi], definitions):
            if a in steal_from_e:
                steal_from_e.pop(a)

        # Remove value just stolen from steal list
        steal_from_e.pop(sym_xi)

    return new_eqs

    # eqs = atomize_eqs(eqs)


def move_forward(atoms):
    reg_usage = get_usage(atoms)
    i = 0
    while i < len(atoms):
        atom = atoms[i]
        killed_regs = 0
        for arg in atom.rhs.atoms():
            if isinstance(arg, Field.Access) or not isinstance(arg, Symbol):
                continue
            reg_usage[arg] -= 1
            if reg_usage[arg] == 0:
                killed_regs += 1
        if killed_regs == 0:
            first_usage = i
            for n in range(i, len(atoms)) or len(
                    [x for x in atoms[n].rhs.atoms() if x in atoms[i].rhs.atoms()]) != 0:
                usage = atoms[n].rhs
                if atom.lhs in usage.atoms():
                    first_usage = n
                    break
            if first_usage - i > 5:
                atoms.insert(first_usage - 1, atoms.pop(i))
                for arg in atom.rhs.atoms():
                    if isinstance(arg, Field.Access) or not isinstance(arg, Symbol):
                        continue
                    reg_usage[arg] += 1
                # print("_move " + str(i) + " " + str(first_usage) + " " +
                #     str(atom))
                i -= 1
        i += 1
    return atoms


def move_backward(atoms):
    reg_usage = get_usage(atoms)
    i = 0
    while i < len(atoms):
        atom = atoms[i]
        killed_regs = 0
        for arg in atom.rhs.atoms():
            if isinstance(arg, Field.Access) or not isinstance(arg, Symbol):
                continue
            reg_usage[arg] -= 1
            if reg_usage[arg] == 0:
                killed_regs += 1
        if killed_regs > 1:
            last_defined = 0
            for n in range(i - 1, 0, -1):
                if len([x for x in atoms[n].rhs.atoms() if x in atoms[i].rhs.atoms()
                        ]) != 0 or atoms[n].lhs in atom.rhs.atoms():
                    last_defined = n
                    break
            if i - last_defined > 5:
                atoms.insert(last_defined + 1, atoms.pop(i))
                # print("_move " + str(i) + " " + str(last_defined) + " " +
                #      str(atom) + " " + str(atoms[last_defined]))
        i += 1

    return atoms


def liveness_analysis(atoms):
    max_alive_regs = 0
    reg_usage = get_usage(atoms)
    alive_atoms = []
    alive_at_peak = []
    for atom in atoms:

        if not isinstance(atom.lhs, Field.Access):
            alive_atoms.append(atom.lhs)
        for arg in atom.rhs.atoms():
            if isinstance(arg, Field.Access) or not isinstance(arg, Symbol):
                continue
            if arg not in alive_atoms:
                print("_referenced Symbol " + str(arg) + " is not alive")
            else:
                reg_usage[arg] -= 1
                if reg_usage[arg] == 0:
                    alive_atoms.remove(arg)
        if max_alive_regs < len(alive_atoms):
            max_alive_regs = len(alive_atoms)
            alive_at_peak = list(alive_atoms)

    #    print("_max alive _registers " + str(max_alive_regs))
    return (alive_at_peak, max_alive_regs)


def schedule_eqs1(eqs):
    definitions = get_definitions(eqs)
    roots = get_roots(eqs)

    def label_eqs(sym, labels):
        if sym not in definitions:
            return (0, 0)
        if sym in labels:
            return labels[sym]
        if isinstance(definitions[sym].rhs, Field.Access):
            labels[sym] = (1, 1)
            return labels[sym]
        reg_counts = []
        for arg in definitions[sym].rhs.atoms():
            reg_counts.append(label_eqs(arg, labels))
        if len(reg_counts) == 1:
            labels[sym] = reg_counts[0]
            return labels[sym]
        print(reg_counts)
        reg_counts.sort(key=lambda x: x[0])
        label = 0
        if reg_counts[-1] == reg_counts[-2]:
            label = reg_counts[-1][0] + 1
        else:
            label = reg_counts[-1][0]
        labels[sym] = (label, 1)
        return labels[sym]

    labels = {}
    for root in roots:
        label_eqs(root.lhs, labels)
    print(labels)

    def schedule_sub_tree(sym, eqs, labels):
        expr = definitions[sym]
        if expr in eqs:
            return
        args = []
        for arg in expr.rhs.atoms():
            if isinstance(arg, Symbol) and not isinstance(arg, Field.Access):
                args.append(arg)
        args.sort(key=lambda arg: -labels[arg][0])
        for arg in args:
            schedule_sub_tree(arg, eqs, labels)
        eqs.append(expr)

    rescheduled_eqs = []
    for root in roots:
        schedule_sub_tree(root.lhs, rescheduled_eqs, labels)

    return rescheduled_eqs


def schedule_eqs2(eqs, target=168, branches=2):
    definitions = get_definitions(eqs)
    roots = get_roots(eqs)

    def recursive_schedule(definitions, needed_syms, usages, target, depth):
        if len(needed_syms) > target:
            return _none
        if len(needed_syms) == 0:
            return []
        sym_list = needed_syms.items()

        score_list = []
        for sym, u in sym_list:
            if u != 0:
                continue
            score = 0
            for arg in definitions[sym].rhs.atoms():
                if arg not in needed_syms and arg in usages:
                    score += 1

            score_list.append((sym, score))
        score_list.sort(key=lambda x: x[1])

        for sym, score in score_list[0:branches]:
            needed_syms.pop(sym)
            for arg in definitions[sym].rhs.atoms():
                if isinstance(arg, Symbol) and not isinstance(arg, Field.Access):
                    if not arg in needed_syms:
                        needed_syms[arg] = usages[arg]
                    needed_syms[arg] -= 1

            instrs = recursive_schedule(definitions, needed_syms, usages, target, depth + 1)
            if not instrs is _none:
                instrs.append(definitions[sym])
                return instrs
            else:
                for arg in definitions[sym].rhs.atoms():
                    if isinstance(arg, Symbol) and not isinstance(arg, Field.Access):
                        needed_syms[arg] += 1
                        if needed_syms[arg] == usages[arg]:
                            needed_syms.pop(arg)
                needed_syms[sym] = 0
        return _none

    usages = get_usage(eqs)
    needed_syms = {u: 0 for u in roots}
    instrs = recursive_schedule(definitions, needed_syms, usages, target, 0)
    return instrs


def schedule_eqs3(eqs, peak_alive=[]):
    peak_alive_set = set(peak_alive)
    # random.shuffle(eqs)
    definitions = get_definitions(eqs)
    definition_atoms = {}
    for sym, definition in definitions.items():
        definition_atoms[sym] = list(definition.rhs.atoms(Symbol))
    roots = get_roots(eqs)
    initial_usages = get_usage(eqs)

    #    levels = [set([frozenset(roots)])]
    level = 0
    current_level_set = set([frozenset(roots)])
    current_usages = {frozenset(roots): {u: 0 for u in roots}}
    current_schedules = {frozenset(roots): (0, [])}
    max_regs = 0
    while len(current_level_set) > 0:
        new_usages = dict()
        new_schedules = dict()
        new_level_set = set()
        min_regs = min([len(current_usages[dec_set]) for dec_set in current_level_set])
        max_regs = max(max_regs, min_regs)
        candidates = [(dec_set, len(current_usages[dec_set]) +
                       len(peak_alive_set.union(set(current_usages[dec_set].keys()))) * 0.1)
                      for dec_set in current_level_set]
        candidates.sort(key=lambda d: d[1])

        for dec_set, regs in candidates[:40]:
            for dec in dec_set:
                new_dec_set = set(dec_set)
                new_dec_set.remove(dec)
                usage = dict(current_usages[dec_set])
                usage.pop(dec)
                atoms = definition_atoms[dec]
                for arg in atoms:
                    if not isinstance(arg, Field.Access):
                        argu = usage.get(arg, initial_usages[arg]) - 1
                        if argu == 0:
                            new_dec_set.add(arg)
                        usage[arg] = argu
                frozen_new_dec_set = frozenset(new_dec_set)
                schedule = current_schedules[dec_set]
                max_reg_count = max(len(usage), schedule[0])
                if frozen_new_dec_set not in new_schedules or max_reg_count < new_schedules[
                    frozen_new_dec_set][0]:
                    new_schedule = list(schedule[1])
                    new_schedule.append(definitions[dec])
                    new_schedules[frozen_new_dec_set] = (max_reg_count, new_schedule)

                if len(frozen_new_dec_set) > 0:
                    new_level_set.add(frozen_new_dec_set)
                new_usages[frozen_new_dec_set] = usage

        current_schedules = new_schedules
        current_usages = new_usages
        current_level_set = new_level_set
        # print(len(current_level_set))
        level += 1

    schedule = current_schedules[frozenset()]
    schedule[1].reverse()
    return (schedule[1])


split_symbol_iter = numbered_symbols("split_")


def split_live_range(eqs, splits=2):
    for i in range(0, splits):
        max_alive_regs = 0
        reg_usage = get_usage(eqs)
        definitions = get_definitions(eqs)
        alive_atoms = []
        alive_at_peak = []
        usage_at_peak = []

        for atom in eqs:
            if not isinstance(atom.lhs, Field.Access):
                alive_atoms.append(atom.lhs)
            for arg in atom.rhs.atoms():
                if isinstance(
                        arg, Field.Access) or not isinstance(arg, Symbol) or arg not in alive_atoms:
                    continue

                else:
                    reg_usage[arg] -= 1
                    if reg_usage[arg] == 0:
                        alive_atoms.remove(arg)
            if max_alive_regs < len(alive_atoms):
                max_alive_regs = len(alive_atoms)
                alive_at_peak = list(alive_atoms)
                usage_at_peak = {u: reg_usage[u] for u in alive_at_peak}
                peak_eq = atom
        peak_idx = eqs.index(peak_eq)
        for sym in alive_at_peak:
            dependent = False
            for arg in definitions[sym].rhs.atoms():
                if arg in definitions:
                    dependent = True
            if dependent:
                continue
            next_occurence = 0
            for i in range(peak_idx, len(eqs)):
                if sym == eqs[i].lhs or sym in eqs[i].rhs.atoms():
                    next_occurence = i
                    break
            new_sym = next(split_symbol_iter)
            if next_occurence == 0:
                continue
            eqs.insert(next_occurence, Assignment(new_sym, definitions[sym].rhs))

            for i in range(peak_idx, len(eqs)):
                if sym in eqs[i].rhs.atoms():
                    eqs[i] = eqs[i].subs(sym, new_sym)

            break
    return eqs


def duplicate_trivial_ops(eqs, nonTrivialLength=3, trivialSymbolLength=1):
    definitions = get_definitions(eqs)
    eq_list = list(eqs)
    idx = 0
    while (idx < len(eq_list)):
        eq = eq_list[idx]
        if isinstance(eq.lhs, Field.Access):
            idx += 1
            continue

        trivial = True
        if len(eq.rhs.atoms()) >= nonTrivialLength:
            trivial = False

        for arg in eq.rhs.atoms():
            if arg in definitions:
                trivial = False
                break

        if len(eq.rhs.atoms(Symbol)) == trivialSymbolLength:
            trivial = True

        if trivial:
            for i in range(idx + 1, len(eq_list)):
                if eq.lhs in eq_list[i].rhs.atoms():
                    eq_list[i] = Assignment(eq_list[i].lhs, eq_list[i].rhs.subs({eq.lhs: eq.rhs}))

            eq_list.remove(eq)
        else:
            idx += 1
    return eq_list


def scramble_eqs(eqs, attempts=1000):
    max_alive_regs = 0
    reg_usage = get_usage(eqs)
    alive_atoms = []
    alive_at_eq = {}
    for atom in eqs:
        alive_at_eq[atom] = {u: reg_usage[u] for u in alive_atoms}

        if not isinstance(atom.lhs, Field.Access):
            alive_atoms.append(atom.lhs)
        for arg in atom.rhs.atoms():
            if isinstance(arg, Field.Access) or not isinstance(arg, Symbol):
                continue
            if arg not in alive_atoms:
                print("_referenced Symbol " + str(arg) + " is not alive")
            else:
                reg_usage[arg] -= 1
                if reg_usage[arg] == 0:
                    alive_atoms.remove(arg)
        if max_alive_regs < len(alive_atoms):
            max_alive_regs = len(alive_atoms)

    orig_usage = get_usage(eqs)
    for i in range(0, attempts):
        a = random.randint(0, len(eqs) - 10)
        if a + 2 >= len(eqs) - 10: continue
        b = random.randint(a + 2, min(len(eqs) - 10, a + 20))
        eqa = eqs[a]
        eqb = eqs[b]
        if max_alive_regs - len(alive_at_eq[eqs[a]]) < 15:
            continue

        if alive_at_eq[eqs[a + 1]][eqa.lhs] == alive_at_eq[eqb].get(eqa.lhs, 0):

            #            print()
            #           print(eqa)
            #          print(eqb)

            usage = alive_at_eq[eqs[a - 1]]
            eqs.insert(b - 1, eqs.pop(a))
            for n in range(a - 1, b + 1):
                atom = eqs[n]
                # print(str(n) + " " + str(atom))
                alive_at_eq[atom] = dict(usage)
                if not isinstance(atom.lhs, Field.Access):
                    usage[atom.lhs] = orig_usage[atom.lhs]
                for arg in atom.rhs.atoms():
                    if isinstance(arg, Field.Access) or not isinstance(arg, Symbol):
                        continue
                    if arg not in usage:
                        pass
                    else:
                        usage[arg] -= 1
                        if usage[arg] == 0:
                            usage.pop(arg)
    return eqs


def scheduling_iteration(eqs):
    atomized_eqs = three_operand_form(eqs)
    eqs = duplicate_trivial_ops(eqs)
    rescheduled_eqs = schedule_eqs3(eqs)
    alive_at_peak = liveness_analysis(rescheduled_eqs)

    for n in range(0, 5):
        split_live_range(atomized_eqs)
    rescheduled_eqs = schedule_eqs3(rescheduled_eqs, alive_at_peak)
    for i in range(0, 10):
        scramble_eqs(atomized_eqs)
    return atomized_eqs


def fuse_subs(eqs):
    new_eqs = copy.copy(eqs)
    for eq in eqs:
        if isinstance(eq.rhs, sympy.Mul) and len(eq.rhs.args) == 2:
            if eq.rhs.args[0] == -1 or eq.rhs.args[1] == -1:
                for i, new_eq in enumerate(new_eqs):
                    if eq.lhs in new_eqs[i].atoms():
                        new_eqs[i] = Assignment(new_eqs[i].lhs, new_eq.rhs.subs(eq.lhs, eq.rhs))
                new_eqs.remove(eq)
    return new_eqs


def fuse_f_m_as(eqs, max_usage=1):
    usage = get_usage(eqs)
    new_eqs = copy.copy(eqs)
    for eq in eqs:
        if isinstance(eq.rhs, sympy.Mul) and len(eq.rhs.args) == 2 and usage[eq.lhs] <= max_usage:
            for index, new_eq in enumerate(new_eqs):
                if isinstance(new_eq.rhs, sympy.Add) and eq.lhs in new_eq.rhs.atoms():
                    no_mul_in_args = True
                    for arg in new_eq.rhs.args:
                        if isinstance(arg, sympy.Mul):
                            no_mul_in_args = False
                    if no_mul_in_args:
                        new_eqs[index] = new_eqs[index].subs(eq.lhs, eq.rhs)
                        usage[eq.lhs] -= 1
                        if usage[eq.lhs] == 0:
                            new_eqs.remove(eq)
                    break

    return new_eqs


def remove_sqrt(input_eqs):
    def remove_in_expr(expr):
        if len(expr.args) > 0:
            if expr.func == sympy._pow and (expr.args[0] == 0.5 or expr.args[1] == 0.5):
                return sympy.Mul(*[remove_in_expr(a) for a in expr.args])
            else:
                return expr.func(*[remove_in_expr(a) for a in expr.args])
        else:
            return expr

    eqs = []
    for eq in input_eqs:
        eqs.append(remove_in_expr(eq))
    return eqs


def remove_div(input_eqs):
    def remove_in_expr(expr):
        if len(expr.args) > 0:
            if expr.func == sympy._pow and (expr.args[0] == -1 or expr.args[1] == -1):
                return sympy.Mul(*[remove_in_expr(a) for a in expr.args])
            else:
                return expr.func(*[remove_in_expr(a) for a in expr.args])
        else:
            return expr

    eqs = []
    for eq in input_eqs:
        eqs.append(remove_in_expr(eq))
    return eqs


def remove_piecewise(input_eqs):
    def remove_in_expr(expr):
        if len(expr.args) > 0:
            if expr.func == sympy.Piecewise:
                cond = 1.0
                summands = []
                for a in expr.args:
                    #     print(remove_in_expr(a[0]))
                    #     print(remove_in_expr(a[1]))
                    #     print( sympy.Mul(cond, remove_in_expr(a[0]), remove_in_expr(a[1])))
                    summands.append(sympy.Mul(cond, remove_in_expr(a[0]), remove_in_expr(a[1])))
                    cond = cond * (1 - remove_in_expr(a[1]))
                #    print()

                #               print(sympy.Add(*summands))
                #               print("----")
                return sympy.Add(*summands)
            else:
                return expr.func(*[remove_in_expr(a) for a in expr.args])
        elif expr == sympy.true:
            return 1.0
        elif expr == sympy.false:
            return 0.0
        else:
            return expr

    eqs = []
    for eq in input_eqs:
        eqs.append(remove_in_expr(eq))
    return eqs


def option_none(eqs):
    return eqs


def option_atomize(eqs):
    return three_operand_form(eqs)


def option_reschedule(eqs):
    return schedule_eqs(eqs)


def option_reschedule_shmem(eqs):
    return schedule_eqs(var_to_shmem(eqs, 8))


def option_liveness_opt_transformation(eqs):
    return liveness_opt_transformation(eqs)


def liveness_opt_transformation_shmem(eqs):
    return var_to_shmem(duplicate_trivial_ops(schedule_eqs(eqs, 20)), 6)


def liveness_opt_transformation_shmem2(eqs):
    return scramble_eqs(schedule_eqs(duplicate_trivial_ops(var_to_shmem(eqs, 6)), 40))


def option_liveness_opt_transformation_shmem(eqs):
    return liveness_opt_transformation_shmem(eqs)


def option_liveness_opt_transformation_shmem2(eqs):
    return liveness_opt_transformation_shmem2(eqs)


def option_atomize_reschedule_no_sqrt(eqs):
    cse_atomized = remove_sqrt(three_operand_form(eqs))
    return schedule_eqs(cse_atomized)


def option_atomize_reschedule_no_div(eqs):
    cse_atomized = remove_div(three_operand_form(eqs))
    return schedule_eqs(cse_atomized)


def option_atomize_reschedule_no_piecewise(eqs):
    cse_atomized = remove_piecewise(three_operand_form(eqs))
    return schedule_eqs(cse_atomized)


def option_atomize_reschedule_no_sqrt_div(eqs):
    cse_atomized = remove_div(remove_sqrt(three_operand_form(eqs)))
    return schedule_eqs(cse_atomized)


def option_atomize_reschedule_no_all(eqs):
    cse_atomized = remove_piecewise(remove_div(remove_sqrt(three_operand_form(eqs))))
    return schedule_eqs(cse_atomized)


def option_atomize_reschedule(eqs):
    return schedule_eqs(three_operand_form(eqs))


def option_reschedule_atomize(eqs):
    return three_operand_form(schedule_eqs(eqs))


def option_reschedule_atomize_scramble(eqs):
    eqs = three_operand_form(schedule_eqs(eqs))
    for i in range(0, 10):
        scramble_eqs(eqs)
    return eqs


def option_dupl_reschedule(eqs):
    eqs = duplicate_trivial_ops(eqs)
    return schedule_eqs(eqs)


def option_dupl_atomize_reschedule(eqs):
    eqs = duplicate_trivial_ops(eqs)
    return schedule_eqs(three_operand_form(eqs))


def option_dupl_atomize_refuse_reschedule(eqs):
    eqs = duplicate_trivial_ops(eqs)
    eqs = fuse_eqs(three_operand_form(eqs))
    return schedule_eqs(eqs)


def option_dupl_reschedule_atomize(eqs):
    eqs = duplicate_trivial_ops(eqs)
    return three_operand_form(schedule_eqs(eqs))


def option_dupl_reschedule_atomize_scramble(eqs):
    eqs = duplicate_trivial_ops(eqs)
    eqs = three_operand_form(schedule_eqs(eqs))
    for i in range(0, 10):
        scramble_eqs(eqs)
    return eqs


def option_sched_iteration(eqs):
    return scheduling_iteration(eqs)


def option_fuse_subs(eqs):
    atomized_eqs = three_operand_form(eqs)
    fused_eqs = fuse_subs(eqs)
    return schedule_eqs(fused_eqs)


def option_fuse_f_m_as(eqs):
    atomized_eqs = three_operand_form(eqs)
    fused_eqs = fuse_f_m_as(eqs)
    return schedule_eqs(fused_eqs)


def option_fuse_both(eqs):
    atomized_eqs = three_operand_form(eqs)
    fused_eqs = fuse_f_m_as(fuse_subs(eqs))
    return schedule_eqs(fused_eqs)


all_sched_options = []
# all_sched_options.append(option_none)
# all_sched_options.append(option_atomize)
# all_sched_options.append(option_reschedule)
# all_sched_options.append(option_atomize_reschedule)
# all_sched_options.append(option_fuse_subs)
# all_sched_options.append(option_fuse_f_m_as)
# all_sched_options.append(option_fuse_both)
# all_sched_options.append(option_reschedule_atomize)
# all_sched_options.append(option_reschedule_atomize_scramble)
# all_sched_options.append(option_dupl_atomize_reschedule)
# all_sched_options.append(option_dupl_atomize_refuse_reschedule)
# all_sched_options.append(option_dupl_reschedule_atomize)
# all_sched_options.append(option_dupl_reschedule_atomize_scramble)
# all_sched_options.append(option_sched_iteration)
# all_sched_options.append(option_atomize_reschedule_no_piecewise)
# all_sched_options.append(option_atomize_reschedule_no_sqrt)
# all_sched_options.append(option_atomize_reschedule_no_div)
# all_sched_options.append(option_atomize_reschedule_no_sqrt_div)
# all_sched_options.append(option_atomize_reschedule_no_all)
# all_sched_options.append(option_liveness_opt_transformation)
all_sched_options.append(option_reschedule_shmem)
all_sched_options.append(option_liveness_opt_transformation_shmem)
all_sched_options.append(option_liveness_opt_transformation_shmem2)


def replace_accesses(eqs):
    access_vars = {}
    for eq in eqs:
        for atom in eq.rhs.atoms():
            if isinstance(atom, Field.Access) and not atom in access_vars:
                access_vars[atom] = Dummy()

    new_eqs = []
    for access in access_vars:
        new_eqs.append(Assignment(access_vars[access], access))

    def replace_accesses_in_expr(expr):
        if len(expr.args) == 0:
            if isinstance(expr, Field.Access):
                return access_vars[expr]
            else:
                return expr
        else:
            return expr.func(*[replace_accesses_in_expr(a) for a in expr.args])

    for eq in eqs:
        new_eqs.append(Assignment(eq.lhs, replace_accesses_in_expr(eq.rhs)))

    return new_eqs


def shifted_equations():
    pass


def get_used_nodes(eqs):
    used_nodes = {}
    for eq in eqs:
        used_nodes[eq.lhs] = []
        for atom in eq.rhs.atoms(Symbol):
            used_nodes[eq.lhs].append(atom)
    return used_nodes


def get_used_by(eqs):
    used_by = {}
    for eq in eqs:
        for arg in eq.rhs.atoms():
            used_by.setdefault(arg, []).append(eq.lhs)
    return used_by

# shifted_equations = copy.deepcopy(equations)

# def shift_accesses(expr):
#    new_args = list(expr.args)
#    print(expr.args)
#    for i in range(0, len(new_args)):
#        if isinstance(new_args[i], Field.Access):
#            new_args[i] = new_args[i].get_shifted(2, 0, 0)
#        if isinstance(new_args[i], sp._expr):
#            shift_accesses(new_args[i])
#    expr._args = new_args
#    print(expr.args)

# for eq in shifted_equations:
#    shift_accesses(eq.rhs)

# for lhs in equations:
#    for rhs in shifted_equations:
#        if lhs == rhs:
#            print(str(lhs.lhs) + " " + str(rhs.lhs))
