# coding: utf-8

# In[32]:


import pickle
import warnings
import pystencils as ps
from pygrandchem.grandchem import GrandChemGenerator
from pygrandchem.scenarios import system_4_2, system_3_1
from pygrandchem.initialization import init_boxes, smooth_fields
from pygrandchem.scenarios import benchmark_configs

from sympy import Number, Symbol, Expr, preorder_traversal, postorder_traversal, Function, Piecewise, relational
from pystencils.simp import sympy_cse_on_assignment_list
from pystencils.simp.liveness_opts import *
from pystencils.simp.liveness_opts_exp import *

from pystencils.simp.liveness_permutations import *


import pycuda

import sys
from subprocess import run, PIPE

from pystencils import show_code
import pycuda.driver as drv


configs = benchmark_configs()

def get_config(name):
    return configs[name]


domain_size = (512, 512, 128)
periodicity = (True, True, False)

optimization = {'gpu_indexing_params': {"block_size": (32, 4, 2)}}



if len(sys.argv) < 4:
    print("Kernel, config and div/sqrt arguments are required")
    exit()

config = get_config(sys.argv[2])
print ("optimizing for config " + sys.argv[2])

approx = False
if sys.argv[3] == "true":
    approx = True


phases, components = config['Parameters']['phases'], config['Parameters']['components']
format_args = {'p': phases, 'c': components, 's': ','.join(str(e) for e in domain_size)}

# Adding fields
dh = ps.create_data_handling(domain_size, periodicity=periodicity, default_target='gpu')
f = dh.fields
phi_src = dh.add_array(
    'phi_src', values_per_cell=config['Parameters']['phases'], layout='fzyx', latex_name='phi_s')
mu_src = dh.add_array(
    'mu_src', values_per_cell=config['Parameters']['components'], layout='fzyx', latex_name="mu_s")
mu_stag = dh.add_array(
    'mu_stag', values_per_cell=(dh.dim, config['Parameters']['components']), layout='f')
phi_stag = dh.add_array('phi_stag', values_per_cell=(dh.dim, phases), layout='f')

phi_dst = dh.add_array_like('phi_dst', 'phi_src')
mu_dst = dh.add_array_like('mu_dst', 'mu_src')

gc = GrandChemGenerator(
    phi_src,
    phi_dst,
    mu_src,
    mu_dst,
    config['FreeEnergy'],
    config['Parameters'],
    #conc=c,
    mu_staggered=mu_stag,
    phi_staggered=phi_stag,
    use_block_offsets=False,
    compile_kernel=False,
    fast_divisions=approx,
    fast_sqrts=approx)

mu_full_eqs = gc.mu_full()
phi_full_eqs = gc.phi_full()


staggered_params = None


if sys.argv[1] == "phi_full":
    eqs = phi_full_eqs

elif sys.argv[1] == "mu_full":
    eqs = mu_full_eqs
elif sys.argv[1] == "mu_partial1":
    staggered_params = gc.mu_partial1()
elif sys.argv[1] == "mu_partial2":
    eqs = gc.mu_partial2()
elif sys.argv[1] == "phi_partial1":
    staggered_params = gc.phi_partial1()
elif sys.argv[1] == "phi_partial2":
    eqs = gc.phi_partial2()
else:
    print("Specified kernel does not exist")
    exit()


if not staggered_params is None:
    eqs = unpack_staggered_eqs(*staggered_params)


phi_kernel = ps.create_kernel(phi_full_eqs, target='gpu', **optimization).compile()
mu_kernel = ps.create_kernel(mu_full_eqs, target='gpu', **optimization).compile()

c = dh.add_array('c', values_per_cell=config['Parameters']['components'], layout='fzyx', gpu=False)

init_boxes(dh)
#initialize_concentration_field(dh, free_energy, config['Parameters']['initial_concentration'])
smooth_fields(dh, sigma=0.4, iterations=5, dim=dh.dim)
dh.synchronization_function(['phi_src', 'phi_dst', 'mu_src', 'mu_dst'])()
print(dh)


def bench_kernels(mu_kernel, phi_kernel):

    start = drv.Event()
    end = drv.Event()

    dh.run_kernel(mu_kernel, timestep=1)
    start.record()
    dh.run_kernel(mu_kernel, timestep=1)
    dh.run_kernel(mu_kernel, timestep=1)
    end.record()
    end.synchronize()
    msec = start.time_till(end) / 2
    print("mu_kernel: {}  {:5.3f} ms".format(mu_kernel.num_regs, msec))

    dh.run_kernel(phi_kernel, timestep=1)
    start.record()
    dh.run_kernel(phi_kernel, timestep=1)
    dh.run_kernel(phi_kernel, timestep=1)
    end.record()
    end.synchronize()
    msec = start.time_till(end) / 2
    print("phi_kernel: {}  {:5.3f} ms".format(phi_kernel.num_regs, msec))



print("warmup")
bench_kernels(mu_kernel, phi_kernel)
dh.swap('mu_src', 'mu_dst')
dh.swap('phi_src', 'phi_dst')







bestSeqs = pickle.load(open('best_seq.pickle', 'rb'))

#bestSeqs= {}
#bestSeqs[( "mu_full", "42_fixT")] =  livenessOptSequence([ [schedule_eqs, [12]], [duplicate_trivial_ops, [1, 1]], [refuse_eqs, [0, 1]],  [var_to_shmem_lt, [3]]],  (64, 4, 2))

#bestSeqs[("phi_full", "42_fixT")] = livenessOptSequence([[schedule_eqs, [34]]],  (16, 2, 2))




if (sys.argv[1], sys.argv[2], sys.argv[3]) in bestSeqs:
    pop = [bestSeqs[(sys.argv[1], sys.argv[2], sys.argv[3])]]
elif (sys.argv[1], sys.argv[2]) in bestSeqs:
    pop = [bestSeqs[(sys.argv[1], sys.argv[2])]]
else:
    pop = [livenessOptSequence()]

print("Loaded seq: " + str(pop[0]) + "\n")




while True:
    pop = evolvePopulation(pop, [eqs], dh, staggered_params)
    bench_kernels(mu_kernel, phi_kernel)

    bestSeqs = pickle.load(open('best_seq.pickle', 'rb'))
    bestSeqs[(sys.argv[1], sys.argv[2], sys.argv[3])] = pop[0]
    pickle.dump(bestSeqs, open('best_seq.pickle', 'wb'))
    print()

